@@ -1,13 +1,13 @@
-import os, sys, io, re, errno, yaml, json, requests, time
-from shutil import copyfile
+import os, sys, io, re, errno, yaml, json, requests, time, shutil
 import datetime as dt
 import pandas as pd
 import numpy as np
-#from sqlalchemy.engine import create_engine
-from sqlalchemy import create_engine
+# --  Teradata Drivers:
+import teradata  # odbc driver
+import sqlalchemy
+from teradataml.context import context as tdml_context
+from teradataml.dataframe import dataframe as tdml_df
 from teradataml.dataframe.copy_to import copy_to_sql
-from teradataml.context.context import *
-from teradataml import *
 
 class tdcoa():
     """DESCRIPTION:
@@ -40,9 +40,6 @@ class tdcoa():
       >>> c.prepare_sql()
       >>> c.execute_run()          # target system VPN
       >>> c.upload_to_transcend()  # transcend VPN
-
-
-
     """
 
     # paths
@@ -52,6 +49,7 @@ class tdcoa():
     secretpath = ''
     filesetpath = ''
     outputpath = ''
+    version = "0.3.2.2"
 
     # log settings
     logs =  []
@@ -59,6 +57,7 @@ class tdcoa():
     bufferlogs = True
     configyaml = ''
     printlog = True
+    show_full_sql = False
     skipgit = False
     skipdbs = False
 
@@ -73,7 +72,7 @@ class tdcoa():
     settings = {}
 
 
-    def __init__(self, approot='.', printlog=True, config='config.yaml', secrets='secrets.yaml', filesets='filesets.yaml'):
+    def __init__(self, approot='.', printlog=True, config='config.yaml', secrets='secrets.yaml', filesets='filesets.yaml', pyfile_only=False):
         self.bufferlog = True
         self.printlog = printlog
         self.approot = os.path.join('.', approot)
@@ -84,20 +83,24 @@ class tdcoa():
         self.log('app root', self.approot)
         self.log('config file', self.configpath)
         self.log('secrets file', self.secretpath)
+        self.log('tdcoa version',  self.version)
 
         self.skipgit = False
         self.skipdbs = False
 
-        # stub in config.yaml, secrets.yaml, if missing
-        if not os.path.isfile(self.configpath):
-            self.log('missing config.yaml', 'creating %s' %self.configpath)
-            self.yaml_config(writefile=True, skipgit=self.skipgit, skipdbs=self.skipdbs)
-        if not os.path.isfile(self.secretpath):
-            self.log('missing secrets.yaml', 'creating %s' %self.secretpath)
-            self.yaml_secrets(writefile=True)
-        # filesets.yaml is validated at download time
+        if pyfile_only:
+            self.pyfile_setup(filename='setup.py')
+        else:
+            # stub in config.yaml, secrets.yaml, if missing
+            if not os.path.isfile(self.configpath):
+                self.log('missing config.yaml', 'creating %s' %self.configpath)
+                self.yaml_config(writefile=True, skipgit=self.skipgit, skipdbs=self.skipdbs)
+            if not os.path.isfile(self.secretpath):
+                self.log('missing secrets.yaml', 'creating %s' %self.secretpath)
+                self.yaml_secrets(writefile=True)
+            # filesets.yaml is validated at download time
 
-        self.reload_config()
+            self.reload_config()
 
 
 
@@ -186,7 +189,6 @@ class tdcoa():
         cy.append('  account:            "Demo Customer"')
         cy.append('  startdate:          "Current_Date - 7"')
         cy.append('  enddate:            "Current_Date - 1"')
-        cy.append('  td_quicklook:       "sh186014"')
         cy.append('  some_value:         "Camel"')
         cy.append('')
         cy.append('')
@@ -239,6 +241,8 @@ class tdcoa():
         cy.append('  username:   "{td_quicklook}"')
         cy.append('  password:   "{td_password}"')
         cy.append('  logmech:    "LDAP"')
+        cy.append('  db_coa:     "adlste_coa"')
+        cy.append('  db_region:  "adlste_westcomm"')
         cy.append('')
         cy.append('')
         cy.append('folders:')
@@ -256,6 +260,10 @@ class tdcoa():
         cy.append('  skip_git:   "%s"' %str(skipgit))
         cy.append('  skip_dbs:   "%s"' %str(skipdbs))
         cy.append('  run_non_fileset_folders: "True" ')
+        cy.append('  customer_connection_type:  "sqlalchemy"')
+        # cy.append('  transcend_connection_type: "teradataml"')
+        cy.append('    # valid connection types: "teradataml", "sqlalchemy", or an odbc driver name')
+
         rtn = '\n'.join(cy)
         if writefile:
             with open( os.path.join(self.configpath), 'w') as fh:
@@ -265,6 +273,51 @@ class tdcoa():
 
 
 
+    def pyfile_setup(self, filename=''):
+        if filename=='': filename = os.path.join('..','setup.py')
+        self.log('\ngenerating setup.py for pypi')
+        cy=[]
+        cy.append('import setuptools ')
+        cy.append('')
+        cy.append('with open("README.md", "r") as fh: ')
+        cy.append('    long_description = fh.read() ')
+        cy.append('')
+        cy.append('setuptools.setup( ')
+        cy.append('    name="tdcsm", ')
+        cy.append('    version="%s", ' %self.version)
+        cy.append('    author="Stephen Hilton", ')
+        cy.append('    author_email="Stephen@FamilyHilton.com", ')
+        cy.append('    description="Teradata tools for CSMs", ')
+        cy.append('    long_description=long_description, ')
+        cy.append('    long_description_content_type="text/markdown", ')
+        cy.append('    url="https://github.com/tdcoa/tdcsm", ')
+        cy.append('    packages=setuptools.find_packages(), ')
+        cy.append('    install_requires=[ ')
+        cy.append('          "pandas", ')
+        cy.append('          "numpy", ')
+        cy.append('          "requests", ')
+        cy.append('          "pyyaml", ')
+        cy.append('          "teradatasqlalchemy", ')
+        cy.append('          "teradataml", ')
+        cy.append('          "teradata" ')
+        cy.append('      ], ')
+        cy.append('    classifiers=[ ')
+        cy.append('        "Programming Language :: Python :: 3", ')
+        cy.append('        "License :: OSI Approved :: MIT License", ')
+        cy.append('        "Operating System :: OS Independent", ')
+        cy.append('    ], ')
+        cy.append('    python_requires=">=3.6", ')
+        cy.append(') ')
+
+        rtn = '\n'.join(cy)
+        with open( filename, 'w') as fh:
+            fh.write(rtn)
+        self.log('  saving file', filename)
+        return None
+
+
+
+
     def sql_create_temp_from_csv(self, csvfilepath):
         tbl = os.path.basename(csvfilepath)
         self.log('    transcribing sql', tbl)
@@ -488,18 +541,20 @@ class tdcoa():
 
 
     def recursively_delete_subfolders(self, parentpath):
+        self.bufferlogs=True
         self.log('purge all subfolders',parentpath)
         for itm in os.listdir(parentpath):
             if os.path.isdir(os.path.join(parentpath,itm)):
-                self.log(' recursively deleting', itm)
-                os.popen('rm -r %s' %os.path.join(parentpath, itm))
-                time.sleep(.333)  # give os time to catch-up
+                self.recursive_delete(os.path.join(parentpath, itm))
+        self.bufferlogs = False
+
 
     def recursive_delete(self, delpath):
         if os.path.isdir(delpath):
             self.log(' recursively deleting', delpath)
-            os.popen('rm -r %s' %delpath)
-            time.sleep(.333)  # give os time to catch-up
+            shutil.rmtree(delpath)
+        else:
+            self.log(' path not found', delpath)
 
 
 
@@ -528,7 +583,7 @@ class tdcoa():
 
                         if os.path.isfile(srcpath):
                             self.log('    file copied', dstpath)
-                            copyfile(srcpath, dstpath)
+                            shutil.copyfile(srcpath, dstpath)
                         elif os.path.isdir(os.path.join(sourcepath,itm)):
                             self.log('    folder copied', dstpath)
                             os.mkdir(dstpath)
@@ -537,6 +592,133 @@ class tdcoa():
                             self.log('    um... unknown filetype: %s' %srcpath)
 
 
+
+    def close_connection(self, connobject, skip=False): # TODO
+        self.log('CLOSE_CONNECTION called',  str(dt.datetime.now()))
+        self.log('*** THIS FUNCTION IS NOT YET IMPLEMENTED ***')
+        conntype = connobject['type']
+        conn = connobject['connection']
+        self.log('  connection type', conntype)
+
+        if skip:
+            self.log('skip dbs setting is true, emulating closure...')
+
+        else:
+            # ------------------------------------
+            if conntype == 'teradataml':
+                pass
+
+            # ------------------------------------
+            elif conntype == 'sqlalchemy':
+                pass
+
+            # ------------------------------------
+            else:  # assume odbc connect
+                pass
+
+        self.log('connection closed', str(dt.datetime.now()))
+        return True
+
+
+
+
+    def open_connection(self, conntype, host='', logmech='', username='', password='', system={}, skip=False):
+        self.log('OPEN_CONNECTION started',  str(dt.datetime.now()))
+        self.log('  connection type', conntype)
+        # check all variables... use system{} as default, individual variables as overrides
+        host=host.strip().lower()
+        logmech=logmech.strip().lower()
+        conntype=conntype.strip().lower()
+
+        if host=='': host=system['host']
+        if logmech=='': logmech=system['logmech']
+        if username=='': username=system['username']
+        if password=='': password=system['password']
+
+        self.log('  host', host)
+        self.log('  logmech', logmech)
+        self.log('  username', username)
+        self.log('  password', password)
+
+        connObject ={'type':conntype, 'skip':skip, 'connection':None, 'components':{}}
+        connObject['components']={'host':host, 'logmech':logmech,
+                                  'username':username, 'password':password}
+
+        self.log('connecting...')
+
+        if skip:
+            self.log('skip dbs setting is true, emulating connection...')
+
+        else:
+            # ------------------------------------
+            if conntype == 'teradataml':
+                if logmech=='': logmech='TD2'
+                connObject['connection'] = tdml_context.create_context(
+                                            host=host,
+                                            logmech=logmech,
+                                            username = username,
+                                            password = password)
+
+            # ------------------------------------
+            elif conntype == 'sqlalchemy':
+                if logmech.strip() != '': logmech = '/?logmech=%s' %logmech
+                connstring = 'teradatasql://%s:%s@%s%s' %(username,password,host,logmech)
+                connObject['connection'] = sqlalchemy.create_engine(connstring)
+
+            # ------------------------------------
+            else:  # assume odbc connect
+                self.log('  (odbc driver)')
+                udaExec = teradata.UdaExec(appName='tdcoa',
+                                           version=self.version,
+                                           logConsole=False)
+                connObject['connection'] = udaExec.connect(method = 'odbc',
+                                           system = host,
+                                           username = username,
+                                           password = password,
+                                           driver = conntype)
+
+        self.log('connected!', str(dt.datetime.now()))
+        return connObject
+
+
+
+
+    def open_sql(self, connobject, sql, skip=False ):
+        import pandas as pd
+        conntype = connobject['type']
+        conn = connobject['connection']
+
+        self.log('connection type', conntype)
+        self.log('sql, first 100 characters:\n  %s' %sql[:100].replace('\n',' ').strip() + '...')
+        self.log('sql submitted', str(dt.datetime.now()))
+
+        if self.show_full_sql:
+            self.log('full sql:', '\n%s\n' %sql)
+
+        if skip:
+            self.log('skip dbs setting is true, emulating execution...')
+            df = pd.DataFrame(columns=list('ABCD'))
+
+        else:
+            # ------------------------------------
+            if conntype == 'teradataml':
+                df = tdml_df.DataFrame.from_query(sql)
+                df = df.to_pandas()
+
+            # ------------------------------------
+            elif conntype == 'sqlalchemy':
+                df = pd.read_sql(sql, conn)
+
+            # ------------------------------------
+            else:  # assume odbc connect
+                df = pd.read_sql(sql, conn)
+
+        self.log('sql completed', str(dt.datetime.now()))
+        self.log('record count', str(len(df)))
+        return df
+
+
+
 # -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 # -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 # -=-=  end utilities
@@ -590,6 +772,7 @@ class tdcoa():
         self.bufferlogs = True
         self.log('reload_config started', header=True)
         self.log('time',str(dt.datetime.now()))
+        self.log('tdcoa version',  self.version)
 
         # load secrets
         self.log('loading secrets', os.path.basename(self.secretpath))
@@ -617,31 +800,34 @@ class tdcoa():
         self.log('loading dictionary', 'transcend')
         self.transcend = configyaml['transcend']
         self.check_setting(self.transcend,
-                           required_item_list=['username','password','host','logmech'],
-                           defaults=['{td_quicklook}','{td_password}','"tdprdcop3.td.teradata.com"','LDAP'])
+                           required_item_list=['username','password','host','logmech','db_coa','db_region'],
+                           defaults=['{td_quicklook}','{td_password}','tdprdcop3.td.teradata.com','TD2','adlste_coa','adlste_westcomm'])
         self.transcend['connectionstring'] = 'teradatasql://%s:%s@%s/?logmech=%s' %(
                                                     self.transcend['username'],
                                                     self.transcend['password'],
                                                     self.transcend['host'],
                                                     self.transcend['logmech'])
+
         # check and set required Folders
         self.log('loading dictionary', 'folders')
         self.folders = configyaml['folders']
         self.check_setting(self.folders, required_item_list=['download','sql','run','output'],
-                           defaults=['"1_download"','"2_sql_store"','"3_ready_to_run"','"4_output"'])
+                           defaults=['1_download','2_sql_store','3_ready_to_run','4_output'])
 
         # check and set required Settings
         self.log('loading dictionary', 'settings')
         self.settings = configyaml['settings']
         self.check_setting(self.settings,
-                           required_item_list=['githost','gitfileset','gitmotd','localfilesets','skip_git','skip_dbs','run_non_fileset_folders'],
-                           defaults=['"https://raw.githubusercontent.com/tdcoa/sql/master/"'
-                                    ,'"filesets.yaml"'
-                                    ,'"motd.txt"'
-                                    ,'"{download}/filesets.yaml"'
-                                    ,'"False"'
-                                    ,'"False"'
-                                    ,'"True"'])
+                           required_item_list=['githost','gitfileset','gitmotd','localfilesets','skip_git','skip_dbs'
+                                              ,'run_non_fileset_folders','customer_connection_type'],
+                           defaults=['https://raw.githubusercontent.com/tdcoa/sql/master/'
+                                    ,'filesets.yaml'
+                                    ,'motd.txt'
+                                    ,'{download}/filesets.yaml'
+                                    ,'False'
+                                    ,'False'
+                                    ,'True'
+                                    ,'sqlalchemy'])
         self.filesetpath = self.settings['localfilesets']
 
         if 'skip_git' in self.settings and str(self.settings['skip_git']).lower() in['true']:
@@ -946,6 +1132,10 @@ class tdcoa():
                                         self.log('  always use dictionary')
                                         runfiletext = self.substitute(runfiletext, self.substitutions, subname='overall app defaults (config.substitutions)')
 
+                                        # SUBSTITUTE values for: TRANSCEND (mostly for db_coa and db_region)
+                                        runfiletext = self.substitute(runfiletext, self.transcend, subname='overall transcend database defaults (db_coa and db_region)',
+                                                                      skipkeys=['host','username','password','logmech'])
+
                                         # SUBSTITUTE values for:   fileset defaults
                                         sub_dict = self.filesets[setfolder]
                                         if self.dict_active(sub_dict, 'fileset defaults'):
@@ -1054,7 +1244,7 @@ class tdcoa():
         prepared sql is stored after the prepare_sql() function.  This includes the
         runlog.txt.  All  files are moved, leaving the 'run' folder empty after the
         operation. The destination folder is a new time-stamped output folder (with
-        optional name).  Useful when you don't have access to execute_sql() (the
+        optional name).  Useful when you don't have access to execute_run() (the
         process that normally archives collateral) against customer system directly,
         but still want to keep a record of that 'run'.   For example, if you need to
         prepare sql to send to a customer DBA for execution - you cannot execute_run()
@@ -1098,10 +1288,14 @@ class tdcoa():
         return outputpath
 
 
+
     def execute_run(self, name=''):
         self.log('execute_run started', header=True)
         self.log('time',str(dt.datetime.now()))
 
+        # TODO: paramterize final database lcoation (adlste_wetcomm should be {}
+        #  when building out the upload_manifest.json, so EMEA and APAC can use
+
         # at this point, we make the assumption that everything in the "run" directory is valid
 
         # make output directory for execution output and other collateral
@@ -1118,125 +1312,124 @@ class tdcoa():
         for sysname in os.listdir(runpath):
             sysfolder = os.path.join(runpath, sysname)
             if os.path.isdir(sysfolder):
+
+                # iterate system folders  -- must exist in config.yaml!
                 if sysname not in self.systems:
                     self.log('SYSTEM NOT FOUND IN CONFIG.YAML', sysname, warning=True)
                 else:
 
+                    # iterate file set folders -- ok to NOT exist, depending on setting
                     for setname in os.listdir(sysfolder):
                         setfolder = os.path.join(sysfolder, setname)
                         if os.path.isdir(setfolder):
 
-                            self.log('SYSTEM:  %s   FILESET:  %s' %(sysname,setname), header=True)
-
-                            workpath = setfolder
-                            outputfo = os.path.join(outputpath, sysname, setname)
-
-                            self.log('work (sql) path', workpath)
-                            self.log('output path', outputfo)
-
-                            # collect all prepared sql files, place in alpha order
-                            coasqlfiles=[]
-                            for coafile in os.listdir(workpath):
-                                if coafile[:1] != '.' and coafile[-8:]=='.coa.sql' :
-                                    self.log('found prepared sql file', coafile )
-                                    coasqlfiles.append(coafile)
-                            coasqlfiles.sort()
-
-                            if len(coasqlfiles)==0:
-                                self.log('no sql files found', warning=True)
+                            if setname not in self.systems[sysname]['filesets'] and str(self.settings['run_non_fileset_folders']).strip().lower() != 'true':
+                                self.log('-'* self.logspace)
+                                self.log('WARNING!!!\nfileset does not exist', setname)
+                                self.log(' AND setting "run_non_fileset_folders" is not "True"')
+                                self.log(' Skipping folder', setname)
                             else:
 
-                                self.log('all sql files alpha-sorted for exeuction consistency')
-                                self.log('sql files found', len(coasqlfiles))
+                                self.log('SYSTEM:  %s   FILESET:  %s' %(sysname,setname), header=True)
+                                workpath = setfolder
+                                outputfo = os.path.join(outputpath, sysname, setname)
 
-                                # create output folder:
-                                self.log('output folder', outputfo)
-                                if not os.path.exists(outputfo):
-                                    os.makedirs(outputfo)
-                                self.outputpath = outputfo
+                                self.log('work (sql) path', workpath)
+                                self.log('output path', outputfo)
 
-                                # create our upload-manifest:
-                                self.log('creating upload manifest file')
-                                with open(os.path.join(outputfo,'upload-manifest.json'),'w') as manifest:
-                                    manifest.write('{"entries":[ ')
-                                manifestdelim='\n '
+                                # collect all prepared sql files, place in alpha order
+                                coasqlfiles=[]
+                                for coafile in os.listdir(workpath):
+                                    if coafile[:1] != '.' and coafile[-8:]=='.coa.sql' :
+                                        self.log('found prepared sql file', coafile )
+                                        coasqlfiles.append(coafile)
+                                coasqlfiles.sort()
 
-                                # connect to customer system:
-                                connstring = self.systems[sysname]['connectionstring']
-                                self.log('CONNECTING TO', connstring)
-                                if self.skipdbs:
-                                    self.log('skipdbs = True', 'emulating connection')
+                                if len(coasqlfiles)==0:
+                                    self.log('no .coa.sql files found in\n  %s' %workpath, warning=True)
                                 else:
-                                    conn = create_engine(connstring) # <------------------------------- Connect to the database
 
+                                    self.log('all sql files alpha-sorted for exeuction consistency')
+                                    self.log('sql files found', len(coasqlfiles))
 
-                                for coasqlfile in sorted(coasqlfiles):  # loop thru all sql files:
-                                    self.log('\nOPENING SQL FILE', coasqlfile)
-                                    with open(os.path.join(workpath, coasqlfile), 'r') as coasqlfilehdlr:
-                                        sqls = coasqlfilehdlr.read()
+                                    # create output folder:
+                                    self.log('output folder', outputfo)
+                                    if not os.path.exists(outputfo):
+                                        os.makedirs(outputfo)
+                                    self.outputpath = outputfo
 
-                                    sqlcnt = 0
-                                    for sql in sqls.split(';'):  # loop thru the sql in the files
-                                        sqlcnt +=1
+                                    # create our upload-manifest:
+                                    self.log('creating upload manifest file')
+                                    with open(os.path.join(outputfo,'upload-manifest.json'),'w') as manifest:
+                                        manifest.write('{"entries":[ ')
+                                    manifestdelim='\n '
 
-                                        if sql.strip() == '':
-                                            self.log('null statement, skipping')
-                                        else:
-                                            self.log('----')
-                                            self.log('execute sql %i' %sqlcnt, sql[:50].replace('\n',' ').strip() + '...')
+                                    # connect to customer system:
+                                    conn = self.open_connection(
+                                                    conntype = self.settings['customer_connection_type'],
+                                                    skip = self.skipdbs,
+                                                    system = self.systems[sysname]) # <------------------------------- Connect to the database
+                                    # loop thru all sql files:
+                                    for coasqlfile in sorted(coasqlfiles):
+                                        self.log('\nOPENING SQL FILE', coasqlfile)
+                                        with open(os.path.join(workpath, coasqlfile), 'r') as coasqlfilehdlr:
+                                            sqls = coasqlfilehdlr.read()
 
-                                            # pull out any embedded SQLcommands:
-                                            sqlcmd = self.get_special_commands(sql)
+                                        sqlcnt = 0
+                                        for sql in sqls.split(';'):  # loop thru the sql in the files
+                                            sqlcnt +=1
 
-                                            # sqlformatted = sqlcmd['sql']
-                                            # del sqlcmd['sql']
-                                            # sql = sqlformatted.replace('\n',' ')
-                                            # while '  ' in sql:
-                                            #     sql = sql.replace('\n',' ').replace('  ',' ')
-
-
-                                            self.log('sql submitted', str(dt.datetime.now()))
-                                            if self.skipdbs:
-                                                self.log('skipdbs = True', 'emulating execution')
-                                                df = pd.DataFrame(columns=list('ABCD'))
+                                            if sql.strip() == '':
+                                                self.log('null statement, skipping')
                                             else:
-                                                df = pd.read_sql(sql, conn)# <------------------------------- Run SQL
-                                            self.log('sql completed', str(dt.datetime.now()))
-                                            self.log('record count', str(len(df)))
 
+                                                self.log('\n---- SQL #%i' %sqlcnt)
 
-                                            if len(df) != 0:  # Save non-empty returns to .csv
+                                                # pull out any embedded SQLcommands:
+                                                sqlcmd = self.get_special_commands(sql)
+                                                sql = sqlcmd.pop('sql','')
 
-                                                if len(sqlcmd)==0:
-                                                    self.log('no special commands found')
+                                                df = self.open_sql(conn, sql, skip=self.skipdbs)# <------------------------------- Run SQL
 
-                                                if 'save' not in sqlcmd:
-                                                    sqlcmd['save'] = '%s.%s--%s' %(sysname, setname, coasqlfile) + '%04d' %sqlcnt + '.csv'
+                                                if len(df) != 0:  # Save non-empty returns to .csv
 
-                                                # once built, append output folder, SiteID on the front, iterative counter if duplicates
-                                                csvfile = os.path.join(outputfo, sqlcmd['save'])
-                                                i=0
-                                                while os.path.isfile(csvfile):
-                                                    i +=1
-                                                    if i==1:
-                                                        csvfile = csvfile[:-4] + '.%03d' %i + csvfile[-4:]
-                                                    else:
-                                                        csvfile = csvfile[:-8] + '.%03d' %i + csvfile[-4:]
-                                                self.log('CSV save location', csvfile)
+                                                    if len(sqlcmd)==0:
+                                                        self.log('no special commands found')
 
-                                                self.log('saving file...')
-                                                df.to_csv(csvfile) # <---------------------- Save to .csv
-                                                self.log('file saved!')
+                                                    if 'save' not in sqlcmd:
+                                                        sqlcmd['save'] = '%s.%s--%s' %(sysname, setname, coasqlfile) + '%04d' %sqlcnt + '.csv'
 
-                                                if 'load' in sqlcmd:  # add to manifest
-                                                    self.log('file marked for loading to Transcend, adding to upload-manifest.json')
-                                                    if 'call' not in sqlcmd:   sqlcmd['call']=''
-                                                    manifest_entry = '%s{"file": "%s",  "table": "%s",  "call": "%s"}' %(manifestdelim, sqlcmd['save'], sqlcmd['load'], sqlcmd['call'])
-                                                    manifestdelim='\n,'
+                                                    # once built, append output folder, SiteID on the front, iterative counter if duplicates
+                                                    csvfile = os.path.join(outputfo, sqlcmd['save'])
+                                                    i=0
+                                                    while os.path.isfile(csvfile):
+                                                        i +=1
+                                                        if i==1:
+                                                            csvfile = csvfile[:-4] + '.%03d' %i + csvfile[-4:]
+                                                        else:
+                                                            csvfile = csvfile[:-8] + '.%03d' %i + csvfile[-4:]
+                                                    self.log('CSV save location', csvfile)
 
-                                                    with open(os.path.join(outputfo,'upload-manifest.json'),'a') as manifest:
-                                                        manifest.write(manifest_entry)
-                                                        self.log('Manifest updated', str(manifest_entry).replace(',',',\n'))
+                                                    self.log('saving file...')
+                                                    df.to_csv(csvfile) # <---------------------- Save to .csv
+                                                    self.log('file saved!')
+
+                                                    if 'load' in sqlcmd:  # add to manifest
+                                                        self.log('file marked for loading to Transcend, adding to upload-manifest.json')
+                                                        if 'call' not in sqlcmd:   sqlcmd['call']=''
+                                                        manifest_entry = '%s{"file": "%s",  "table": "%s",  "call": "%s"}' %(manifestdelim, sqlcmd['save'], sqlcmd['load'], sqlcmd['call'])
+                                                        manifestdelim='\n,'
+
+                                                        with open(os.path.join(outputfo,'upload-manifest.json'),'a') as manifest:
+                                                            manifest.write(manifest_entry)
+                                                            self.log('Manifest updated', str(manifest_entry).replace(',',',\n'))
+
+                                        # archive file we just processed (for re-run-ability)
+                                        self.log('Moving coa.sql file to Output folder', coasqlfile)
+                                        src = os.path.join(workpath, coasqlfile)
+                                        dst = os.path.join(outputfo, coasqlfile)
+                                        shutil.move(src, dst)
+                                        self.log('')
 
                                 # close JSON object
                                 with open(os.path.join(outputfo,'upload-manifest.json'),'a') as manifest:
@@ -1244,8 +1437,8 @@ class tdcoa():
                                     self.log('closing out upload-manifest.json')
 
                                 # Move all files from run folder to output, for posterity:
-                                self.log('moving all run artifacts to output folder, for archiving')
-                                self.recursive_copy(workpath, outputfo, True)
+                                self.log('moving all other run artifacts to output folder, for archiving')
+                                self.recursive_copy(workpath, outputfo, replace_existing=False)
                                 self.recursive_delete(workpath)
 
 
@@ -1256,7 +1449,7 @@ class tdcoa():
                      self.configpath, self.filesetpath]:
             self.log('copy to output folder root, for ease of use: \n  %s' %srcpath)
             dstpath = os.path.join(outputpath, os.path.basename(srcpath))
-            copyfile(srcpath, dstpath)
+            shutil.copyfile(srcpath, dstpath)
 
         self.log('\ndone!')
         self.log('time', str(dt.datetime.now()))
@@ -1264,9 +1457,7 @@ class tdcoa():
         # after logging is done, move the log file too...
         runlogsrc = os.path.join(self.approot, self.folders['run'], 'runlog.txt')
         runlogdst = os.path.join(outputpath, 'runlog.txt')
-        copyfile(runlogsrc, runlogdst)
-        os.remove(runlogsrc)
-
+        shutil.move(runlogsrc, runlogdst)
 
 
 
@@ -1312,17 +1503,14 @@ class tdcoa():
         self.log('    logmech', self.transcend['logmech'])
         self.log('    username', self.transcend['username'])
         self.log('    password', self.transcend['password'])
+        self.log('    db_coa', self.transcend['db_coa'])
         self.log("\nNOTE:  if you happen to see a scary WARNING below, DON'T PANIC!")
         self.log("       it just means you already had an active connection that was replaced.\n")
-        if self.skipdbs:
-            self.log('skipdbs = True', 'emulating connection')
-        else:
-            transcend = create_context(host = self.transcend['host'],
-                                       username = self.transcend['username'],
-                                       password = self.transcend['password'],
-                                       logmech = self.transcend['logmech'] )  # <--------- Connect
-        self.log('connected!')
 
+        transcend = self.open_connection(
+                        'teradataml',
+                        system = self.transcend,
+                        skip = self.skipdbs)   # <--------------------------------- Connect
 
         # Walk the directory structure looking for upload_manifest.json files
         for workpath, subfo, files in os.walk(outputpath):
@@ -1410,12 +1598,8 @@ class tdcoa():
                                 if self.skipdbs:
                                     self.log('skipdbs = True', 'emulating call')
                                 else:
-                                    transcend.execute('call %s ;' %str(entry['call']) )
+                                    transcend['connection'].execute('call %s ;' %str(entry['call']) )
                                 self.log('complete', str(dt.datetime.now()))
 
-
-        if not self.skipdbs:
-            remove_context()
-
         self.log('\ndone!')
         self.log('time', str(dt.datetime.now()))
